@@ -8,14 +8,14 @@
  * Copyright (c) 2004 Joe English
  */
 
-#include <string.h>
-#include <stdio.h>
-#include <tkInt.h>
-#include <X11/Xatom.h>
-
+#include "tkInt.h"
 #include "ttkTheme.h"
 #include "ttkWidget.h"
 
+#ifdef _WIN32
+#include "tkWinInt.h"
+#endif
+
 /*
  * Extra bits for core.flags:
  */
@@ -147,12 +147,12 @@ typedef struct {
 /*
  * Default option values:
  */
-#define DEF_SELECT_BG	"#000000"
-#define DEF_SELECT_FG	"#ffffff"
-#define DEF_INSERT_BG	"black"
-#define DEF_ENTRY_WIDTH	"20"
-#define DEF_ENTRY_FONT	"TkTextFont"
-#define DEF_LIST_HEIGHT	"10"
+#define DEF_SELECT_BG		"#000000"
+#define DEF_SELECT_FG		"#FFFFFF"
+#define DEF_INSERT_BG		"black"
+#define DEF_ENTRY_WIDTH		"20"
+#define DEF_ENTRY_FONT		"TkTextFont"
+#define DEF_LIST_HEIGHT		"10"
 
 static Tk_OptionSpec EntryOptionSpecs[] = {
     {TK_OPTION_BOOLEAN, "-exportselection", "exportSelection",
@@ -178,7 +178,7 @@ static Tk_OptionSpec EntryOptionSpecs[] = {
 	TK_OPTION_NULL_OK,0,TEXTVAR_CHANGED},
     {TK_OPTION_STRING_TABLE, "-validate", "validate", "Validate",
 	"none", -1, Tk_Offset(Entry, entry.validate),
-	0, (ClientData) validateStrings, 0},
+	TK_OPTION_ENUM_VAR, validateStrings, 0},
     {TK_OPTION_STRING, "-validatecommand", "validateCommand", "ValidateCommand",
 	NULL, -1, Tk_Offset(Entry, entry.validateCmd),
 	TK_OPTION_NULL_OK, 0, 0},
@@ -287,7 +287,7 @@ static char *EntryDisplayString(const char *showChar, int numChars)
 
     TkUtfToUniChar(showChar, &ch);
     size = TkUniCharToUtf(ch, buf);
-    p = displayString = ckalloc(numChars * size + 1);
+    p = displayString = (char *)ckalloc(numChars * size + 1);
 
     while (numChars--) {
 	memcpy(p, buf, size);
@@ -332,8 +332,8 @@ static int
 EntryFetchSelection(
     ClientData clientData, int offset, char *buffer, int maxBytes)
 {
-    Entry *entryPtr = (Entry *) clientData;
-    size_t byteCount;
+    Entry *entryPtr = (Entry *)clientData;
+    int byteCount;
     const char *string;
     const char *selStart, *selEnd;
 
@@ -343,11 +343,11 @@ EntryFetchSelection(
     }
     string = entryPtr->entry.displayString;
 
-    selStart = Tcl_UtfAtIndex(string, entryPtr->entry.selectFirst);
-    selEnd = Tcl_UtfAtIndex(selStart,
+    selStart = TkUtfAtIndex(string, entryPtr->entry.selectFirst);
+    selEnd = TkUtfAtIndex(selStart,
 	    entryPtr->entry.selectLast - entryPtr->entry.selectFirst);
     byteCount = selEnd - selStart - offset;
-    if (byteCount > (size_t)maxBytes) {
+    if (byteCount > maxBytes) {
     /* @@@POSSIBLE BUG: Can transfer partial UTF-8 sequences.  Is this OK? */
 	byteCount = maxBytes;
     }
@@ -365,7 +365,7 @@ EntryFetchSelection(
  */
 static void EntryLostSelection(ClientData clientData)
 {
-    Entry *entryPtr = (Entry *) clientData;
+    Entry *entryPtr = (Entry *)clientData;
     entryPtr->core.flags &= ~GOT_SELECTION;
     entryPtr->entry.selectFirst = entryPtr->entry.selectLast = -1;
     TtkRedisplayWidget(&entryPtr->core);
@@ -381,7 +381,7 @@ static void EntryOwnSelection(Entry *entryPtr)
 	&& (!Tcl_IsSafe(entryPtr->core.interp))
 	&& !(entryPtr->core.flags & GOT_SELECTION)) {
 	Tk_OwnSelection(entryPtr->core.tkwin, XA_PRIMARY, EntryLostSelection,
-		(ClientData) entryPtr);
+		entryPtr);
 	entryPtr->core.flags |= GOT_SELECTION;
     }
 }
@@ -397,8 +397,8 @@ static void EntryOwnSelection(Entry *entryPtr)
 static void
 ExpandPercents(
      Entry *entryPtr,		/* Entry that needs validation. */
-     const char *template, 	/* Script template */
-     const char *new,		/* Potential new value of entry string */
+     const char *templ, 	/* Script template */
+     const char *newValue,		/* Potential new value of entry string */
      int index,			/* index of insert/delete */
      int count,			/* #changed characters */
      VREASON reason,		/* Reason for change */
@@ -411,28 +411,28 @@ ExpandPercents(
     int ch;
     char numStorage[2*TCL_INTEGER_SPACE];
 
-    while (*template) {
+    while (*templ) {
 	/* Find everything up to the next % character and append it
 	 * to the result string.
 	 */
-	string = Tcl_UtfFindFirst(template, '%');
+	string = Tcl_UtfFindFirst(templ, '%');
 	if (string == NULL) {
 	    /* No more %-sequences to expand.
 	     * Copy the rest of the template.
 	     */
-	    Tcl_DStringAppend(dsPtr, template, -1);
+	    Tcl_DStringAppend(dsPtr, templ, -1);
 	    return;
 	}
-	if (string != template) {
-	    Tcl_DStringAppend(dsPtr, template, string - template);
-	    template = string;
+	if (string != templ) {
+	    Tcl_DStringAppend(dsPtr, templ, string - templ);
+	    templ = string;
 	}
 
 	/* There's a percent sequence here.  Process it.
 	 */
-	++template; /* skip over % */
-	if (*template != '\0') {
-	    template += TkUtfToUniChar(template, &ch);
+	++templ; /* skip over % */
+	if (*templ != '\0') {
+	    templ += TkUtfToUniChar(templ, &ch);
 	} else {
 	    ch = '%';
 	}
@@ -447,26 +447,26 @@ ExpandPercents(
 		} else {
 		    number = -1;
 		}
-		sprintf(numStorage, "%d", number);
+		snprintf(numStorage, sizeof(numStorage), "%d", number);
 		string = numStorage;
 		break;
 	    case 'i': /* index of insert/delete */
-		sprintf(numStorage, "%d", index);
+		snprintf(numStorage, sizeof(numStorage), "%d", index);
 		string = numStorage;
 		break;
 	    case 'P': /* 'Peeked' new value of the string */
-		string = new;
+		string = newValue;
 		break;
 	    case 's': /* Current string value */
 		string = entryPtr->entry.string;
 		break;
 	    case 'S': /* string to be inserted/deleted, if any */
 		if (reason == VALIDATE_INSERT) {
-		    string = Tcl_UtfAtIndex(new, index);
-		    stringLength = Tcl_UtfAtIndex(string, count) - string;
+		    string = TkUtfAtIndex(newValue, index);
+		    stringLength = TkUtfAtIndex(string, count) - string;
 		} else if (reason == VALIDATE_DELETE) {
-		    string = Tcl_UtfAtIndex(entryPtr->entry.string, index);
-		    stringLength = Tcl_UtfAtIndex(string, count) - string;
+		    string = TkUtfAtIndex(entryPtr->entry.string, index);
+		    stringLength = TkUtfAtIndex(string, count) - string;
 		} else {
 		    string = "";
 		    stringLength = 0;
@@ -506,9 +506,9 @@ ExpandPercents(
 static int RunValidationScript(
     Tcl_Interp *interp, 	/* Interpreter to use */
     Entry *entryPtr,		/* Entry being validated */
-    const char *template,	/* Script template */
+    const char *templ,	/* Script template */
     const char *optionName,	/* "-validatecommand", "-invalidcommand" */
-    const char *new,		/* Potential new value of entry string */
+    const char *newValue,	/* Potential new value of entry string */
     int index,			/* index of insert/delete */
     int count,			/* #changed characters */
     VREASON reason)		/* Reason for change */
@@ -517,7 +517,7 @@ static int RunValidationScript(
     int code;
 
     Tcl_DStringInit(&script);
-    ExpandPercents(entryPtr, template, new, index, count, reason, &script);
+    ExpandPercents(entryPtr, templ, newValue, index, count, reason, &script);
     code = Tcl_EvalEx(interp,
 		Tcl_DStringValue(&script), Tcl_DStringLength(&script),
 		TCL_EVAL_GLOBAL);
@@ -561,7 +561,7 @@ static int EntryNeedsValidation(VMODE vmode, VREASON reason)
  * Returns:
  *	TCL_OK if the change is accepted
  *	TCL_BREAK if the change is rejected
- *      TCL_ERROR if any errors occured
+ *	TCL_ERROR if any errors occurred
  *
  * The change will be rejected if -validatecommand returns 0,
  * or if -validatecommand or -invalidcommand modifies the value.
@@ -578,9 +578,9 @@ EntryValidateChange(
     VMODE vmode = entryPtr->entry.validate;
     int code, change_ok;
 
-    if (   (entryPtr->entry.validateCmd == NULL)
+    if ((entryPtr->entry.validateCmd == NULL)
 	|| (entryPtr->core.flags & VALIDATING)
-	|| !EntryNeedsValidation(vmode, reason) )
+	|| !EntryNeedsValidation(vmode, reason))
     {
 	return TCL_OK;
     }
@@ -634,7 +634,10 @@ done:
  * Returns:
  * 	TCL_OK if valid, TCL_BREAK if invalid, TCL_ERROR on error.
  */
-static int EntryRevalidate(Tcl_Interp *interp, Entry *entryPtr, VREASON reason)
+static int EntryRevalidate(
+    TCL_UNUSED(Tcl_Interp *),
+    Entry *entryPtr,
+    VREASON reason)
 {
     int code = EntryValidateChange(
 		    entryPtr, entryPtr->entry.string, -1,0, reason);
@@ -654,8 +657,12 @@ static int EntryRevalidate(Tcl_Interp *interp, Entry *entryPtr, VREASON reason)
 static void EntryRevalidateBG(Entry *entryPtr, VREASON reason)
 {
     Tcl_Interp *interp = entryPtr->core.interp;
-    if (EntryRevalidate(interp, entryPtr, reason) == TCL_ERROR) {
-	Tcl_BackgroundException(interp, TCL_ERROR);
+    VMODE vmode = entryPtr->entry.validate;
+
+    if (EntryNeedsValidation(vmode, reason)) {
+        if (EntryRevalidate(interp, entryPtr, reason) == TCL_ERROR) {
+	    Tcl_BackgroundException(interp, TCL_ERROR);
+        }
     }
 }
 
@@ -725,7 +732,7 @@ EntryStoreValue(Entry *entryPtr, const char *value)
 
     /* Store new value:
      */
-    entryPtr->entry.string = ckalloc(numBytes + 1);
+    entryPtr->entry.string = (char *)ckalloc(numBytes + 1);
     strcpy(entryPtr->entry.string, value);
     entryPtr->entry.numBytes = numBytes;
     entryPtr->entry.numChars = numChars;
@@ -782,7 +789,7 @@ static int EntrySetValue(Entry *entryPtr, const char *value)
  */
 static void EntryTextVariableTrace(void *recordPtr, const char *value)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
 
     if (WidgetDestroyed(&entryPtr->core)) {
 	return;
@@ -812,33 +819,33 @@ InsertChars(
     const char *value)		/* New characters to add */
 {
     char *string = entryPtr->entry.string;
-    size_t byteIndex = Tcl_UtfAtIndex(string, index) - string;
+    size_t byteIndex = TkUtfAtIndex(string, index) - string;
     size_t byteCount = strlen(value);
     int charsAdded = Tcl_NumUtfChars(value, byteCount);
     size_t newByteCount = entryPtr->entry.numBytes + byteCount + 1;
-    char *new;
+    char *newBytes;
     int code;
 
     if (byteCount == 0) {
 	return TCL_OK;
     }
 
-    new =  ckalloc(newByteCount);
-    memcpy(new, string, byteIndex);
-    strcpy(new + byteIndex, value);
-    strcpy(new + byteIndex + byteCount, string + byteIndex);
+    newBytes =  (char *)ckalloc(newByteCount);
+    memcpy(newBytes, string, byteIndex);
+    strcpy(newBytes + byteIndex, value);
+    strcpy(newBytes + byteIndex + byteCount, string + byteIndex);
 
     code = EntryValidateChange(
-	    entryPtr, new, index, charsAdded, VALIDATE_INSERT);
+	    entryPtr, newBytes, index, charsAdded, VALIDATE_INSERT);
 
     if (code == TCL_OK) {
 	AdjustIndices(entryPtr, index, charsAdded);
-	code = EntrySetValue(entryPtr, new);
+	code = EntrySetValue(entryPtr, newBytes);
     } else if (code == TCL_BREAK) {
 	code = TCL_OK;
     }
 
-    ckfree(new);
+    ckfree(newBytes);
     return code;
 }
 
@@ -853,37 +860,37 @@ DeleteChars(
 {
     char *string = entryPtr->entry.string;
     size_t byteIndex, byteCount, newByteCount;
-    char *new;
+    char *newBytes;
     int code;
 
     if (index < 0) {
 	index = 0;
     }
-    if (count > entryPtr->entry.numChars - index) {
+    if (count + index > entryPtr->entry.numChars) {
 	count = entryPtr->entry.numChars - index;
     }
     if (count <= 0) {
 	return TCL_OK;
     }
 
-    byteIndex = Tcl_UtfAtIndex(string, index) - string;
-    byteCount = Tcl_UtfAtIndex(string+byteIndex, count) - (string+byteIndex);
+    byteIndex = TkUtfAtIndex(string, index) - string;
+    byteCount = TkUtfAtIndex(string+byteIndex, count) - (string+byteIndex);
 
     newByteCount = entryPtr->entry.numBytes + 1 - byteCount;
-    new =  ckalloc(newByteCount);
-    memcpy(new, string, byteIndex);
-    strcpy(new + byteIndex, string + byteIndex + byteCount);
+    newBytes =  (char *)ckalloc(newByteCount);
+    memcpy(newBytes, string, byteIndex);
+    strcpy(newBytes + byteIndex, string + byteIndex + byteCount);
 
     code = EntryValidateChange(
-	    entryPtr, new, index, count, VALIDATE_DELETE);
+	    entryPtr, newBytes, index, count, VALIDATE_DELETE);
 
     if (code == TCL_OK) {
 	AdjustIndices(entryPtr, index, -count);
-	code = EntrySetValue(entryPtr, new);
+	code = EntrySetValue(entryPtr, newBytes);
     } else if (code == TCL_BREAK) {
 	code = TCL_OK;
     }
-    ckfree(new);
+    ckfree(newBytes);
 
     return code;
 }
@@ -900,7 +907,7 @@ DeleteChars(
 static void
 EntryEventProc(ClientData clientData, XEvent *eventPtr)
 {
-    Entry *entryPtr = (Entry *) clientData;
+    Entry *entryPtr = (Entry *)clientData;
 
     Tcl_Preserve(clientData);
     switch (eventPtr->type) {
@@ -923,17 +930,19 @@ EntryEventProc(ClientData clientData, XEvent *eventPtr)
  */
 
 static void
-EntryInitialize(Tcl_Interp *interp, void *recordPtr)
+EntryInitialize(
+    TCL_UNUSED(Tcl_Interp *),
+    void *recordPtr)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
 
     Tk_CreateEventHandler(
 	entryPtr->core.tkwin, EntryEventMask, EntryEventProc, entryPtr);
     Tk_CreateSelHandler(entryPtr->core.tkwin, XA_PRIMARY, XA_STRING,
-	EntryFetchSelection, (ClientData) entryPtr, XA_STRING);
+	EntryFetchSelection, entryPtr, XA_STRING);
     TtkBlinkCursor(&entryPtr->core);
 
-    entryPtr->entry.string		= ckalloc(1);
+    entryPtr->entry.string		= (char *)ckalloc(1);
     *entryPtr->entry.string 		= '\0';
     entryPtr->entry.displayString	= entryPtr->entry.string;
     entryPtr->entry.textVariableTrace 	= 0;
@@ -952,7 +961,7 @@ EntryInitialize(Tcl_Interp *interp, void *recordPtr)
 static void
 EntryCleanup(void *recordPtr)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
 
     if (entryPtr->entry.textVariableTrace)
 	Ttk_UntraceVariable(entryPtr->entry.textVariableTrace);
@@ -974,12 +983,12 @@ EntryCleanup(void *recordPtr)
  */
 static int EntryConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     Tcl_Obj *textVarName = entryPtr->entry.textVariableObj;
     Ttk_TraceHandle *vt = 0;
 
     if (mask & TEXTVAR_CHANGED) {
-	if (textVarName && *Tcl_GetString(textVarName)) {
+	if (textVarName && *Tcl_GetString(textVarName) != '\0') {
 	    vt = Ttk_TraceVariable(interp,
 		    textVarName,EntryTextVariableTrace,entryPtr);
 	    if (!vt) return TCL_ERROR;
@@ -1038,9 +1047,12 @@ static int EntryConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
 /* EntryPostConfigure --
  * 	Post-configuration hook for entry widgets.
  */
-static int EntryPostConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
+static int EntryPostConfigure(
+    TCL_UNUSED(Tcl_Interp *),
+    void *recordPtr,
+    int mask)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int status = TCL_OK;
 
     if ((mask & TEXTVAR_CHANGED) && entryPtr->entry.textVariableTrace != NULL) {
@@ -1079,7 +1091,7 @@ EntryCharPosition(Entry *entryPtr, int index)
 static void
 EntryDoLayout(void *recordPtr)
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     WidgetCore *corePtr = &entryPtr->core;
     Tk_TextLayout textLayout = entryPtr->entry.textLayout;
     int leftIndex = entryPtr->entry.xscroll.first;
@@ -1155,7 +1167,7 @@ static GC EntryGetGC(Entry *entryPtr, Tcl_Obj *colorObj, TkRegion clip)
 	mask |= GCForeground;
     }
     gc = Tk_GetGC(entryPtr->core.tkwin, mask, &gcValues);
-    if (clip != None) {
+    if (clip != NULL) {
 	TkSetRegion(Tk_Display(entryPtr->core.tkwin), gc, clip);
     }
     return gc;
@@ -1166,7 +1178,7 @@ static GC EntryGetGC(Entry *entryPtr, Tcl_Obj *colorObj, TkRegion clip)
  */
 static void EntryDisplay(void *clientData, Drawable d)
 {
-    Entry *entryPtr = clientData;
+    Entry *entryPtr = (Entry *)clientData;
     Tk_Window tkwin = entryPtr->core.tkwin;
     int leftIndex = entryPtr->entry.xscroll.first,
 	rightIndex = entryPtr->entry.xscroll.last + 1,
@@ -1265,7 +1277,7 @@ static void EntryDisplay(void *clientData, Drawable d)
 	    cursorX = field.x + field.width - cursorWidth;
 	}
 
-	gc = EntryGetGC(entryPtr, es.insertColorObj, None);
+	gc = EntryGetGC(entryPtr, es.insertColorObj, NULL);
 	XFillRectangle(Tk_Display(tkwin), d, gc,
 	    cursorX, cursorY, cursorWidth, cursorHeight);
 	Tk_FreeGC(Tk_Display(tkwin), gc);
@@ -1274,16 +1286,27 @@ static void EntryDisplay(void *clientData, Drawable d)
     /* Draw the text:
      */
     gc = EntryGetGC(entryPtr, es.foregroundObj, clipRegion);
-    Tk_DrawTextLayout(
-	Tk_Display(tkwin), d, gc, entryPtr->entry.textLayout,
-	entryPtr->entry.layoutX, entryPtr->entry.layoutY,
-	leftIndex, rightIndex);
-    XSetClipMask(Tk_Display(tkwin), gc, None);
-    Tk_FreeGC(Tk_Display(tkwin), gc);
-
-    /* Overwrite the selected portion (if any) in the -selectforeground color:
-     */
     if (showSelection) {
+
+        /* Draw the selected and unselected portions separately.
+	 */
+	if (leftIndex < selFirst) {
+	    Tk_DrawTextLayout(
+		Tk_Display(tkwin), d, gc, entryPtr->entry.textLayout,
+		entryPtr->entry.layoutX, entryPtr->entry.layoutY,
+		leftIndex, selFirst);
+	}
+	if (selLast < rightIndex) {
+	    Tk_DrawTextLayout(
+		Tk_Display(tkwin), d, gc, entryPtr->entry.textLayout,
+		entryPtr->entry.layoutX, entryPtr->entry.layoutY,
+		selLast, rightIndex);
+	}
+	XSetClipMask(Tk_Display(tkwin), gc, None);
+	Tk_FreeGC(Tk_Display(tkwin), gc);
+
+	/* Draw the selected portion in the -selectforeground color:
+	 */
 	gc = EntryGetGC(entryPtr, es.selForegroundObj, clipRegion);
 	Tk_DrawTextLayout(
 	    Tk_Display(tkwin), d, gc, entryPtr->entry.textLayout,
@@ -1291,13 +1314,23 @@ static void EntryDisplay(void *clientData, Drawable d)
 	    selFirst, selLast);
 	XSetClipMask(Tk_Display(tkwin), gc, None);
 	Tk_FreeGC(Tk_Display(tkwin), gc);
+    } else {
+
+        /* Draw the entire visible text
+         */
+	Tk_DrawTextLayout(
+	    Tk_Display(tkwin), d, gc, entryPtr->entry.textLayout,
+	    entryPtr->entry.layoutX, entryPtr->entry.layoutY,
+	    leftIndex, rightIndex);
+	XSetClipMask(Tk_Display(tkwin), gc, None);
+	Tk_FreeGC(Tk_Display(tkwin), gc);
     }
 
     /* Drop the region. Note that we have to manually remove the reference to
      * it from the Xft guts (if they're being used).
      */
 #ifdef HAVE_XFT
-    TkUnixSetXftClipRegion(None);
+    TkUnixSetXftClipRegion(NULL);
 #endif
     TkDestroyRegion(clipRegion);
 }
@@ -1366,6 +1399,7 @@ EntryIndex(
 	*indexPtr = Tk_PointToChar(entryPtr->entry.textLayout,
 		x - entryPtr->entry.layoutX, 0);
 
+        TtkUpdateScrollInfo(entryPtr->entry.xscrollHandle);
 	if (*indexPtr < entryPtr->entry.xscroll.first) {
 	    *indexPtr = entryPtr->entry.xscroll.first;
 	}
@@ -1381,7 +1415,7 @@ EntryIndex(
 	    *indexPtr += 1;
 	}
     } else {
-	if (Tcl_GetInt(interp, string, indexPtr) != TCL_OK) {
+	if (Tcl_GetIntFromObj(interp, indexObj, indexPtr) != TCL_OK) {
 	    goto badIndex;
 	}
 	if (*indexPtr < 0) {
@@ -1406,7 +1440,7 @@ static int
 EntryBBoxCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     Ttk_Box b;
     int index;
 
@@ -1436,7 +1470,7 @@ static int
 EntryDeleteCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int first, last;
 
     if ((objc < 3) || (objc > 4)) {
@@ -1465,7 +1499,7 @@ static int
 EntryGetCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     if (objc != 2) {
 	Tcl_WrongNumArgs(interp, 2, objv, NULL);
 	return TCL_ERROR;
@@ -1481,7 +1515,7 @@ static int
 EntryICursorCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     if (objc != 3) {
 	Tcl_WrongNumArgs(interp, 2, objv, "pos");
 	return TCL_ERROR;
@@ -1501,7 +1535,7 @@ static int
 EntryIndexCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int index;
 
     if (objc != 3) {
@@ -1523,7 +1557,7 @@ static int
 EntryInsertCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int index;
 
     if (objc != 4) {
@@ -1545,7 +1579,7 @@ EntryInsertCommand(
 static int EntrySelectionClearCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
 
     if (objc != 3) {
 	Tcl_WrongNumArgs(interp, 3, objv, NULL);
@@ -1562,7 +1596,7 @@ static int EntrySelectionClearCommand(
 static int EntrySelectionPresentCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     if (objc != 3) {
 	Tcl_WrongNumArgs(interp, 3, objv, NULL);
 	return TCL_ERROR;
@@ -1578,13 +1612,13 @@ static int EntrySelectionPresentCommand(
 static int EntrySelectionRangeCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int start, end;
     if (objc != 5) {
 	Tcl_WrongNumArgs(interp, 3, objv, "start end");
 	return TCL_ERROR;
     }
-    if (    EntryIndex(interp, entryPtr, objv[3], &start) != TCL_OK
+    if (EntryIndex(interp, entryPtr, objv[3], &start) != TCL_OK
          || EntryIndex(interp, entryPtr, objv[4], &end) != TCL_OK) {
 	return TCL_ERROR;
     }
@@ -1616,7 +1650,7 @@ static const Ttk_Ensemble EntrySelectionCommands[] = {
 static int EntrySetCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     if (objc != 3) {
 	Tcl_WrongNumArgs(interp, 2, objv, "value");
 	return TCL_ERROR;
@@ -1632,7 +1666,7 @@ static int EntrySetCommand(
 static int EntryValidateCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     int code;
 
     if (objc != 2) {
@@ -1654,13 +1688,13 @@ static int EntryValidateCommand(
 static int EntryXViewCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Entry *entryPtr = recordPtr;
+    Entry *entryPtr = (Entry *)recordPtr;
     if (objc == 3) {
 	int newFirst;
 	if (EntryIndex(interp, entryPtr, objv[2], &newFirst) != TCL_OK) {
 	    return TCL_ERROR;
 	}
-	TtkScrollTo(entryPtr->entry.xscrollHandle, newFirst);
+	TtkScrollTo(entryPtr->entry.xscrollHandle, newFirst, 1);
 	return TCL_OK;
     }
     return TtkScrollviewCommand(interp, objc, objv, entryPtr->entry.xscrollHandle);
@@ -1703,6 +1737,16 @@ static WidgetSpec EntryWidgetSpec = {
     EntryDisplay		/* displayProc */
 };
 
+/*------------------------------------------------------------------------
+ * Named indices for the combobox "current" command
+ */
+static const char *const comboboxCurrentIndexNames[] = {
+    "end", NULL
+};
+enum comboboxCurrentIndices {
+    INDEX_END
+};
+
 /*------------------------------------------------------------------------
  * +++ Combobox widget record.
  */
@@ -1711,7 +1755,7 @@ typedef struct {
     Tcl_Obj	*postCommandObj;
     Tcl_Obj	*valuesObj;
     Tcl_Obj	*heightObj;
-    int 	currentIndex;
+    int	currentIndex;
 } ComboboxPart;
 
 typedef struct {
@@ -1739,7 +1783,7 @@ static Tk_OptionSpec ComboboxOptionSpecs[] = {
 static void
 ComboboxInitialize(Tcl_Interp *interp, void *recordPtr)
 {
-    Combobox *cb = recordPtr;
+    Combobox *cb = (Combobox *)recordPtr;
 
     cb->combobox.currentIndex = -1;
     TtkTrackElementState(&cb->core);
@@ -1752,7 +1796,7 @@ ComboboxInitialize(Tcl_Interp *interp, void *recordPtr)
 static int
 ComboboxConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
 {
-    Combobox *cbPtr = recordPtr;
+    Combobox *cbPtr = (Combobox *)recordPtr;
     int unused;
 
     /* Make sure -values is a valid list:
@@ -1772,13 +1816,13 @@ ComboboxConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
 static int ComboboxCurrentCommand(
     void *recordPtr, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
 {
-    Combobox *cbPtr = recordPtr;
+    Combobox *cbPtr = (Combobox *)recordPtr;
     int currentIndex = cbPtr->combobox.currentIndex;
     const char *currentValue = cbPtr->entry.string;
     int nValues;
     Tcl_Obj **values;
 
-    Tcl_ListObjGetElements(interp,cbPtr->combobox.valuesObj,&nValues,&values);
+    Tcl_ListObjGetElements(interp, cbPtr->combobox.valuesObj, &nValues, &values);
 
     if (objc == 2) {
 	/* Check if currentIndex still valid:
@@ -1804,15 +1848,51 @@ static int ComboboxCurrentCommand(
 	Tcl_SetObjResult(interp, Tcl_NewIntObj(currentIndex));
 	return TCL_OK;
     } else if (objc == 3) {
-	if (Tcl_GetIntFromObj(interp, objv[2], &currentIndex) != TCL_OK) {
-	    return TCL_ERROR;
-	}
-	if (currentIndex < 0 || currentIndex >= nValues) {
-	    Tcl_SetObjResult(interp, Tcl_ObjPrintf(
-		    "Index %s out of range", Tcl_GetString(objv[2])));
-	    Tcl_SetErrorCode(interp, "TTK", "COMBOBOX", "IDX_RANGE", NULL);
-	    return TCL_ERROR;
-	}
+        int result, index;
+
+        result = Tcl_GetIndexFromObj(NULL, objv[2], comboboxCurrentIndexNames,
+                "", 0, &index);
+        if (result == TCL_OK) {
+
+            /*
+             * The index is one of the named indices.
+             */
+
+	    switch (index) {
+	    case INDEX_END:
+	        /* "end" index */
+		if (nValues <= 0) {
+		    Tcl_SetObjResult(interp, Tcl_ObjPrintf(
+			    "index \"end\" out of range"));
+		    Tcl_SetErrorCode(interp, "TTK", "COMBOBOX", "IDX_RANGE", NULL);
+		    return TCL_ERROR;
+		}
+                currentIndex = nValues - 1;
+                break;
+ 	    default:
+		Tcl_Panic("Unknown named index");
+		return TCL_ERROR;
+	    }
+	} else {
+
+            /*
+             * The index should be just an integer.
+             */
+
+	    if (Tcl_GetIntFromObj(NULL, objv[2], &currentIndex) != TCL_OK) {
+	        Tcl_SetObjResult(interp, Tcl_ObjPrintf(
+		        "Incorrect index %s", Tcl_GetString(objv[2])));
+	        Tcl_SetErrorCode(interp, "TTK", "COMBOBOX", "IDX_VALUE", NULL);
+	        return TCL_ERROR;
+	    }
+
+	    if (currentIndex < 0 || currentIndex >= nValues) {
+	        Tcl_SetObjResult(interp, Tcl_ObjPrintf(
+		        "Index %s out of range", Tcl_GetString(objv[2])));
+	        Tcl_SetErrorCode(interp, "TTK", "COMBOBOX", "IDX_RANGE", NULL);
+	        return TCL_ERROR;
+	    }
+        }
 
 	cbPtr->combobox.currentIndex = currentIndex;
 
@@ -1918,7 +1998,7 @@ static Tk_OptionSpec SpinboxOptionSpecs[] = {
 static void
 SpinboxInitialize(Tcl_Interp *interp, void *recordPtr)
 {
-    Spinbox *sb = recordPtr;
+    Spinbox *sb = (Spinbox *)recordPtr;
     TtkTrackElementState(&sb->core);
     EntryInitialize(interp, recordPtr);
 }
@@ -1929,7 +2009,7 @@ SpinboxInitialize(Tcl_Interp *interp, void *recordPtr)
 static int
 SpinboxConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
 {
-    Spinbox *sb = recordPtr;
+    Spinbox *sb = (Spinbox *)recordPtr;
     int unused;
 
     /* Make sure -values is a valid list:
@@ -1991,14 +2071,18 @@ static Ttk_ElementOptionSpec TextareaElementOptions[] = {
 	Tk_Offset(TextareaElement,fontObj), DEF_ENTRY_FONT },
     { "-width", TK_OPTION_INT,
 	Tk_Offset(TextareaElement,widthObj), "20" },
-    { NULL, 0, 0, NULL }
+    { NULL, TK_OPTION_BOOLEAN, 0, NULL }
 };
 
 static void TextareaElementSize(
-    void *clientData, void *elementRecord, Tk_Window tkwin,
-    int *widthPtr, int *heightPtr, Ttk_Padding *paddingPtr)
+    TCL_UNUSED(void *),
+    void *elementRecord,
+    Tk_Window tkwin,
+    int *widthPtr,
+    int *heightPtr,
+    TCL_UNUSED(Ttk_Padding *))
 {
-    TextareaElement *textarea = elementRecord;
+    TextareaElement *textarea = (TextareaElement *)elementRecord;
     Tk_Font font = Tk_GetFontFromObj(tkwin, textarea->fontObj);
     int avgWidth = Tk_TextWidth(font, "0", 1);
     Tk_FontMetrics fm;
@@ -2034,7 +2118,7 @@ TTK_END_LAYOUT
 TTK_BEGIN_LAYOUT(ComboboxLayout)
     TTK_GROUP("Combobox.field", TTK_FILL_BOTH,
 	TTK_NODE("Combobox.downarrow", TTK_PACK_RIGHT|TTK_FILL_Y)
-	TTK_GROUP("Combobox.padding", TTK_FILL_BOTH|TTK_PACK_LEFT|TTK_EXPAND,
+	TTK_GROUP("Combobox.padding", TTK_FILL_BOTH,
 	    TTK_NODE("Combobox.textarea", TTK_FILL_BOTH)))
 TTK_END_LAYOUT
 
