@@ -97,7 +97,7 @@ typedef struct PendingCommand {
 				 * NULL means end of list. */
 } PendingCommand;
 
-typedef struct ThreadSpecificData {
+typedef struct {
     PendingCommand *pendingCommands;
 				/* List of all commands currently being waited
 				 * for. */
@@ -261,12 +261,15 @@ RegOpen(
     unsigned long bytesAfter;
     Atom actualType;
     char **propertyPtr;
+    Tk_ErrorHandler handler;
 
     if (dispPtr->commTkwin == NULL) {
 	SendInit(interp, dispPtr);
     }
 
-    regPtr = ckalloc(sizeof(NameRegistry));
+    handler = Tk_CreateErrorHandler(dispPtr->display, -1, -1, -1, NULL, NULL);
+
+    regPtr = (NameRegistry *)ckalloc(sizeof(NameRegistry));
     regPtr->dispPtr = dispPtr;
     regPtr->locked = 0;
     regPtr->modified = 0;
@@ -306,8 +309,11 @@ RegOpen(
 	XDeleteProperty(dispPtr->display,
 		RootWindow(dispPtr->display, 0),
 		dispPtr->registryProperty);
+        XSync(dispPtr->display, False);
     }
 
+    Tk_DeleteErrorHandler(handler);
+
     /*
      * Xlib placed an extra null byte after the end of the property, just to
      * make sure that it is always NULL-terminated. Be sure to include this
@@ -470,10 +476,10 @@ RegAddName(
     char id[30], *newProp;
     int idLength, newBytes;
 
-    sprintf(id, "%x ", (unsigned) commWindow);
+    snprintf(id, sizeof(id), "%x ", (unsigned) commWindow);
     idLength = strlen(id);
     newBytes = idLength + strlen(name) + 1;
-    newProp = ckalloc(regPtr->propLength + newBytes);
+    newProp = (char *)ckalloc(regPtr->propLength + newBytes);
     strcpy(newProp, id);
     strcpy(newProp+idLength, name);
     if (regPtr->property != NULL) {
@@ -514,6 +520,11 @@ RegClose(
     NameRegistry *regPtr)	/* Pointer to a registry opened with a
 				 * previous call to RegOpen. */
 {
+    Tk_ErrorHandler handler;
+
+    handler = Tk_CreateErrorHandler(regPtr->dispPtr->display, -1, -1, -1,
+            NULL, NULL);
+
     if (regPtr->modified) {
 	if (!regPtr->locked && !localData.sendDebug) {
 	    Tcl_Panic("The name registry was modified without being locked!");
@@ -540,6 +551,8 @@ RegClose(
 
     XFlush(regPtr->dispPtr->display);
 
+    Tk_DeleteErrorHandler(handler);
+
     if (regPtr->property != NULL) {
 	if (regPtr->allocedByX) {
 	    XFree(regPtr->property);
@@ -795,7 +808,7 @@ Tk_SetAppName(
     const char *actualName;
     Tcl_DString dString;
     int offset, i;
-    ThreadSpecificData *tsdPtr =
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
 
     interp = winPtr->mainPtr->interp;
@@ -817,13 +830,13 @@ Tk_SetAppName(
 	     * the "send" command to the interpreter.
 	     */
 
-	    riPtr = ckalloc(sizeof(RegisteredInterp));
+	    riPtr = (RegisteredInterp *)ckalloc(sizeof(RegisteredInterp));
 	    riPtr->interp = interp;
 	    riPtr->dispPtr = winPtr->dispPtr;
 	    riPtr->nextPtr = tsdPtr->interpListPtr;
 	    tsdPtr->interpListPtr = riPtr;
 	    riPtr->name = NULL;
-	    Tcl_CreateCommand(interp, "send", Tk_SendCmd, riPtr, DeleteProc);
+	    Tcl_CreateObjCommand(interp, "send", Tk_SendObjCmd, riPtr, DeleteProc);
 	    if (Tcl_IsSafe(interp)) {
 		Tcl_HideCommand(interp, "send", "send");
 	    }
@@ -862,7 +875,7 @@ Tk_SetAppName(
 		Tcl_DStringSetLength(&dString, offset+TCL_INTEGER_SPACE);
 		actualName = Tcl_DStringValue(&dString);
 	    }
-	    sprintf(Tcl_DStringValue(&dString) + offset, "%d", i);
+	    snprintf(Tcl_DStringValue(&dString) + offset, TCL_INTEGER_SPACE, "%d", i);
 	}
 	w = RegFindName(regPtr, actualName);
 	if (w == None) {
@@ -901,7 +914,7 @@ Tk_SetAppName(
 
     RegAddName(regPtr, actualName, Tk_WindowId(dispPtr->commTkwin));
     RegClose(regPtr);
-    riPtr->name = ckalloc(strlen(actualName) + 1);
+    riPtr->name = (char *)ckalloc(strlen(actualName) + 1);
     strcpy(riPtr->name, actualName);
     if (actualName != name) {
 	Tcl_DStringFree(&dString);
@@ -914,7 +927,7 @@ Tk_SetAppName(
 /*
  *--------------------------------------------------------------
  *
- * Tk_SendCmd --
+ * Tk_SendObjCmd --
  *
  *	This function is invoked to process the "send" Tcl command. See the
  *	user documentation for details on what it does.
@@ -929,27 +942,32 @@ Tk_SetAppName(
  */
 
 int
-Tk_SendCmd(
-    ClientData clientData,	/* Information about sender (only dispPtr
+Tk_SendObjCmd(
+    TCL_UNUSED(void *),	/* Information about sender (only dispPtr
 				 * field is used). */
     Tcl_Interp *interp,		/* Current interpreter. */
-    int argc,			/* Number of arguments. */
-    const char **argv)		/* Argument strings. */
+    int objc,			/* Number of arguments. */
+    Tcl_Obj *const objv[])	/* Argument strings. */
 {
+    enum {
+	SEND_ASYNC, SEND_DISPLAYOF, SEND_LAST
+    };
+    static const char *const sendOptions[] = {
+	"-async",   "-displayof",   "--",  NULL
+    };
     TkWindow *winPtr;
     Window commWindow;
     PendingCommand pending;
-    register RegisteredInterp *riPtr;
+    RegisteredInterp *riPtr;
     const char *destName;
-    int result, c, async, i, firstArg;
-    size_t length;
+    int result, index, async, i, firstArg;
     Tk_RestrictProc *prevProc;
     ClientData prevArg;
     TkDisplay *dispPtr;
     Tcl_Time timeout;
     NameRegistry *regPtr;
     Tcl_DString request;
-    ThreadSpecificData *tsdPtr =
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
     Tcl_Interp *localInterp;	/* Used when the interpreter to send the
 				 * command to is within the same process. */
@@ -963,43 +981,31 @@ Tk_SendCmd(
     if (winPtr == NULL) {
 	return TCL_ERROR;
     }
-    for (i = 1; i < (argc-1); ) {
-	if (argv[i][0] != '-') {
+    for (i = 1; i < objc; i++) {
+	if (Tcl_GetIndexFromObjStruct(NULL, objv[i], sendOptions,
+		sizeof(char *), "option", 0, &index) != TCL_OK) {
 	    break;
 	}
-	c = argv[i][1];
-	length = strlen(argv[i]);
-	if ((c == 'a') && (strncmp(argv[i], "-async", length) == 0)) {
-	    async = 1;
-	    i++;
-	} else if ((c == 'd') && (strncmp(argv[i], "-displayof",
-		length) == 0)) {
-	    winPtr = (TkWindow *) Tk_NameToWindow(interp, argv[i+1],
+	if (index == SEND_ASYNC) {
+	    ++async;
+	} else if (index == SEND_DISPLAYOF) {
+	    winPtr = (TkWindow *) Tk_NameToWindow(interp, Tcl_GetString(objv[++i]),
 		    (Tk_Window) winPtr);
 	    if (winPtr == NULL) {
 		return TCL_ERROR;
 	    }
-	    i += 2;
-	} else if (strcmp(argv[i], "--") == 0) {
+	} else if (index == SEND_LAST) {
 	    i++;
 	    break;
-	} else {
-	    Tcl_SetObjResult(interp, Tcl_ObjPrintf(
-		    "bad option \"%s\": must be -async, -displayof, or --",
-		    argv[i]));
-	    Tcl_SetErrorCode(interp, "TK", "SEND", "OPTION", NULL);
-	    return TCL_ERROR;
 	}
     }
 
-    if (argc < (i+2)) {
-	Tcl_SetObjResult(interp, Tcl_ObjPrintf("wrong # args: should be "
-		"\"%s ?-option value ...? interpName arg ?arg ...?\"",
-		argv[0]));
-	Tcl_SetErrorCode(interp, "TCL", "WRONGARGS", NULL);
+    if (objc < (i+2)) {
+	Tcl_WrongNumArgs(interp, 1, objv,
+		"?-option value ...? interpName arg ?arg ...?");
 	return TCL_ERROR;
     }
-    destName = argv[i];
+    destName = Tcl_GetString(objv[i]);
     firstArg = i+1;
 
     dispPtr = winPtr->dispPtr;
@@ -1023,16 +1029,16 @@ Tk_SendCmd(
 	Tcl_Preserve(riPtr);
 	localInterp = riPtr->interp;
 	Tcl_Preserve(localInterp);
-	if (firstArg == (argc-1)) {
-	    result = Tcl_GlobalEval(localInterp, argv[firstArg]);
+	if (firstArg == (objc-1)) {
+	    result = Tcl_EvalEx(localInterp, Tcl_GetString(objv[firstArg]), -1, TCL_EVAL_GLOBAL);
 	} else {
 	    Tcl_DStringInit(&request);
-	    Tcl_DStringAppend(&request, argv[firstArg], -1);
-	    for (i = firstArg+1; i < argc; i++) {
+	    Tcl_DStringAppend(&request, Tcl_GetString(objv[firstArg]), -1);
+	    for (i = firstArg+1; i < objc; i++) {
 		Tcl_DStringAppend(&request, " ", 1);
-		Tcl_DStringAppend(&request, argv[i], -1);
+		Tcl_DStringAppend(&request, Tcl_GetString(objv[i]), -1);
 	    }
-	    result = Tcl_GlobalEval(localInterp, Tcl_DStringValue(&request));
+	    result = Tcl_EvalEx(localInterp, Tcl_DStringValue(&request), -1, TCL_EVAL_GLOBAL);
 	    Tcl_DStringFree(&request);
 	}
 	if (interp != localInterp) {
@@ -1090,17 +1096,42 @@ Tk_SendCmd(
     if (!async) {
 	char buffer[TCL_INTEGER_SPACE * 2];
 
-	sprintf(buffer, "%x %d",
+	snprintf(buffer, sizeof(buffer), "%x %d",
 		(unsigned) Tk_WindowId(dispPtr->commTkwin),
 		localData.sendSerial);
 	Tcl_DStringAppend(&request, "\0-r ", 4);
 	Tcl_DStringAppend(&request, buffer, -1);
     }
     Tcl_DStringAppend(&request, "\0-s ", 4);
-    Tcl_DStringAppend(&request, argv[firstArg], -1);
-    for (i = firstArg+1; i < argc; i++) {
+    Tcl_DStringAppend(&request, Tcl_GetString(objv[firstArg]), -1);
+    for (i = firstArg+1; i < objc; i++) {
 	Tcl_DStringAppend(&request, " ", 1);
-	Tcl_DStringAppend(&request, argv[i], -1);
+	Tcl_DStringAppend(&request, Tcl_GetString(objv[i]), -1);
+    }
+
+    if (!async) {
+	/*
+	 * Register the fact that we're waiting for a command to complete
+	 * (this is needed by SendEventProc and by AppendErrorProc to pass
+	 * back the command's results). Set up a timeout handler so that
+	 * we can check during long sends to make sure that the destination
+	 * application is still alive.
+	 *
+	 * We prepare the pending struct here in order to catch potential
+	 * early X errors from AppendPropCarefully() due to XSync().
+	 */
+
+	pending.serial = localData.sendSerial;
+	pending.dispPtr = dispPtr;
+	pending.target = destName;
+	pending.commWindow = commWindow;
+	pending.interp = interp;
+	pending.result = NULL;
+	pending.errorInfo = NULL;
+	pending.errorCode = NULL;
+	pending.gotResponse = 0;
+	pending.nextPtr = tsdPtr->pendingCommands;
+	tsdPtr->pendingCommands = &pending;
     }
     (void) AppendPropCarefully(dispPtr->display, commWindow,
 	    dispPtr->commProperty, Tcl_DStringValue(&request),
@@ -1115,26 +1146,6 @@ Tk_SendCmd(
 	return TCL_OK;
     }
 
-    /*
-     * Register the fact that we're waiting for a command to complete (this is
-     * needed by SendEventProc and by AppendErrorProc to pass back the
-     * command's results). Set up a timeout handler so that we can check
-     * during long sends to make sure that the destination application is
-     * still alive.
-     */
-
-    pending.serial = localData.sendSerial;
-    pending.dispPtr = dispPtr;
-    pending.target = destName;
-    pending.commWindow = commWindow;
-    pending.interp = interp;
-    pending.result = NULL;
-    pending.errorInfo = NULL;
-    pending.errorCode = NULL;
-    pending.gotResponse = 0;
-    pending.nextPtr = tsdPtr->pendingCommands;
-    tsdPtr->pendingCommands = &pending;
-
     /*
      * Enter a loop processing X events until the result comes in or the
      * target is declared to be dead. While waiting for a result, look only at
@@ -1164,7 +1175,7 @@ Tk_SendCmd(
 		    msg = "target application died";
 		}
 		pending.code = TCL_ERROR;
-		pending.result = ckalloc(strlen(msg) + 1);
+		pending.result = (char *)ckalloc(strlen(msg) + 1);
 		strcpy(pending.result, msg);
 		pending.gotResponse = 1;
 	    } else {
@@ -1345,7 +1356,7 @@ TkSendCleanup(
 
 static int
 SendInit(
-    Tcl_Interp *interp,		/* Interpreter to use for error reporting (no
+    TCL_UNUSED(Tcl_Interp *),	/* Interpreter to use for error reporting (no
 				 * errors are ever returned, but the
 				 * interpreter is needed anyway). */
     TkDisplay *dispPtr)		/* Display to initialize. */
@@ -1408,14 +1419,14 @@ SendEventProc(
     ClientData clientData,	/* Display information. */
     XEvent *eventPtr)		/* Information about event. */
 {
-    TkDisplay *dispPtr = clientData;
+    TkDisplay *dispPtr = (TkDisplay *)clientData;
     char *propInfo, **propInfoPtr = &propInfo;
     const char *p;
     int result, actualFormat;
     unsigned long numItems, bytesAfter;
     Atom actualType;
     Tcl_Interp *remoteInterp;	/* Interp in which to execute the command. */
-    ThreadSpecificData *tsdPtr =
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
 
     if ((eventPtr->xproperty.atom != dispPtr->commProperty)
@@ -1565,7 +1576,7 @@ SendEventProc(
 	    remoteInterp = riPtr->interp;
 	    Tcl_Preserve(remoteInterp);
 
-	    result = Tcl_GlobalEval(remoteInterp, script);
+	    result = Tcl_EvalEx(remoteInterp, script, -1, TCL_EVAL_GLOBAL);
 
 	    /*
 	     * The call to Tcl_Release may have released the interpreter which
@@ -1575,7 +1586,7 @@ SendEventProc(
 	     */
 
 	    if (commWindow != None) {
-		Tcl_DStringAppend(&reply, Tcl_GetStringResult(remoteInterp),
+		Tcl_DStringAppend(&reply, Tcl_GetString(Tcl_GetObjResult(remoteInterp)),
 			-1);
 		if (result == TCL_ERROR) {
 		    const char *varValue;
@@ -1609,7 +1620,7 @@ SendEventProc(
 		if (result != TCL_OK) {
 		    char buffer[TCL_INTEGER_SPACE];
 
-		    sprintf(buffer, "%d", result);
+		    snprintf(buffer, sizeof(buffer), "%d", result);
 		    Tcl_DStringAppend(&reply, "\0-c ", 4);
 		    Tcl_DStringAppend(&reply, buffer, -1);
 		}
@@ -1687,16 +1698,16 @@ SendEventProc(
 		}
 		pcPtr->code = code;
 		if (resultString != NULL) {
-		    pcPtr->result = ckalloc(strlen(resultString) + 1);
+		    pcPtr->result = (char *)ckalloc(strlen(resultString) + 1);
 		    strcpy(pcPtr->result, resultString);
 		}
 		if (code == TCL_ERROR) {
 		    if (errorInfo != NULL) {
-			pcPtr->errorInfo = ckalloc(strlen(errorInfo) + 1);
+			pcPtr->errorInfo = (char *)ckalloc(strlen(errorInfo) + 1);
 			strcpy(pcPtr->errorInfo, errorInfo);
 		    }
 		    if (errorCode != NULL) {
-			pcPtr->errorCode = ckalloc(strlen(errorCode) + 1);
+			pcPtr->errorCode = (char *)ckalloc(strlen(errorCode) + 1);
 			strcpy(pcPtr->errorCode, errorCode);
 		    }
 		}
@@ -1763,15 +1774,14 @@ AppendPropCarefully(
  * operation above.
  */
 
-	/* ARGSUSED */
 static int
 AppendErrorProc(
     ClientData clientData,	/* Command to mark complete, or NULL. */
-    XErrorEvent *errorPtr)	/* Information about error. */
+    TCL_UNUSED(XErrorEvent *))	/* Information about error. */
 {
-    PendingCommand *pendingPtr = clientData;
-    register PendingCommand *pcPtr;
-    ThreadSpecificData *tsdPtr =
+    PendingCommand *pendingPtr = (PendingCommand *)clientData;
+    PendingCommand *pcPtr;
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
 
     if (pendingPtr == NULL) {
@@ -1785,8 +1795,8 @@ AppendErrorProc(
     for (pcPtr = tsdPtr->pendingCommands; pcPtr != NULL;
 	    pcPtr = pcPtr->nextPtr) {
 	if ((pcPtr == pendingPtr) && (pcPtr->result == NULL)) {
-	    pcPtr->result = ckalloc(strlen(pcPtr->target) + 50);
-	    sprintf(pcPtr->result, "no application named \"%s\"",
+	    pcPtr->result = (char *)ckalloc(strlen(pcPtr->target) + 50);
+	    snprintf(pcPtr->result, strlen(pcPtr->target) + 50, "no application named \"%s\"",
 		    pcPtr->target);
 	    pcPtr->code = TCL_ERROR;
 	    pcPtr->gotResponse = 1;
@@ -1818,10 +1828,10 @@ DeleteProc(
     ClientData clientData)	/* Info about registration, passed as
 				 * ClientData. */
 {
-    RegisteredInterp *riPtr = clientData;
-    register RegisteredInterp *riPtr2;
+    RegisteredInterp *riPtr = (RegisteredInterp *)clientData;
+    RegisteredInterp *riPtr2;
     NameRegistry *regPtr;
-    ThreadSpecificData *tsdPtr =
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
 
     regPtr = RegOpen(riPtr->interp, riPtr->dispPtr, 1);
@@ -1863,11 +1873,10 @@ DeleteProc(
  *----------------------------------------------------------------------
  */
 
-    /* ARGSUSED */
 static Tk_RestrictAction
 SendRestrictProc(
-    ClientData clientData,		/* Not used. */
-    register XEvent *eventPtr)		/* Event that just arrived. */
+    TCL_UNUSED(void *),		/* Not used. */
+    XEvent *eventPtr)		/* Event that just arrived. */
 {
     TkDisplay *dispPtr;
 
@@ -1910,7 +1919,7 @@ UpdateCommWindow(
 {
     Tcl_DString names;
     RegisteredInterp *riPtr;
-    ThreadSpecificData *tsdPtr =
+    ThreadSpecificData *tsdPtr = (ThreadSpecificData *)
 	    Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData));
 
     Tcl_DStringInit(&names);
@@ -1943,49 +1952,63 @@ UpdateCommWindow(
  *----------------------------------------------------------------------
  */
 
-	/* ARGSUSED */
 int
 TkpTestsendCmd(
     ClientData clientData,	/* Main window for application. */
     Tcl_Interp *interp,		/* Current interpreter. */
-    int argc,			/* Number of arguments. */
-    const char **argv)		/* Argument strings. */
+    int objc,			/* Number of arguments. */
+    Tcl_Obj *const objv[])		/* Argument strings. */
 {
-    TkWindow *winPtr = clientData;
+    enum {
+	TESTSEND_BOGUS, TESTSEND_PROP, TESTSEND_SERIAL
+    };
+    static const char *const testsendOptions[] = {
+	"bogus",   "prop",   "serial",  NULL
+    };
+    TkWindow *winPtr = (TkWindow *)clientData;
+    Tk_ErrorHandler handler;
+    int index;
 
-    if (argc < 2) {
-	Tcl_AppendResult(interp, "wrong # args; must be \"", argv[0],
-		" option ?arg ...?\"", NULL);
+    if (objc < 2) {
+	Tcl_WrongNumArgs(interp, 1, objv,
+		"option ?arg ...?");
 	return TCL_ERROR;
     }
 
-    if (strcmp(argv[1], "bogus") == 0) {
+    if (Tcl_GetIndexFromObjStruct(interp, objv[1], testsendOptions,
+		sizeof(char *), "option", 0, &index) != TCL_OK) {
+	return TCL_ERROR;
+    }
+    if (index == TESTSEND_BOGUS) {
+        handler = Tk_CreateErrorHandler(winPtr->dispPtr->display, -1, -1, -1,
+                NULL, NULL);
 	XChangeProperty(winPtr->dispPtr->display,
 		RootWindow(winPtr->dispPtr->display, 0),
 		winPtr->dispPtr->registryProperty, XA_INTEGER, 32,
 		PropModeReplace,
 		(unsigned char *) "This is bogus information", 6);
-    } else if (strcmp(argv[1], "prop") == 0) {
+        Tk_DeleteErrorHandler(handler);
+    } else if (index == TESTSEND_PROP) {
 	int result, actualFormat;
 	unsigned long length, bytesAfter;
 	Atom actualType, propName;
 	char *property, **propertyPtr = &property, *p, *end;
 	Window w;
 
-	if ((argc != 4) && (argc != 5)) {
-	    Tcl_AppendResult(interp, "wrong # args; must be \"", argv[0],
-		    " prop window name ?value ?\"", NULL);
+	if ((objc != 4) && (objc != 5)) {
+		Tcl_WrongNumArgs(interp, 1, objv,
+			"prop window name ?value ?");
 	    return TCL_ERROR;
 	}
-	if (strcmp(argv[2], "root") == 0) {
+	if (strcmp(Tcl_GetString(objv[2]), "root") == 0) {
 	    w = RootWindow(winPtr->dispPtr->display, 0);
-	} else if (strcmp(argv[2], "comm") == 0) {
+	} else if (strcmp(Tcl_GetString(objv[2]), "comm") == 0) {
 	    w = Tk_WindowId(winPtr->dispPtr->commTkwin);
 	} else {
-	    w = strtoul(argv[2], &end, 0);
+	    w = strtoul(Tcl_GetString(objv[2]), &end, 0);
 	}
-	propName = Tk_InternAtom((Tk_Window) winPtr, argv[3]);
-	if (argc == 4) {
+	propName = Tk_InternAtom((Tk_Window) winPtr, Tcl_GetString(objv[3]));
+	if (objc == 4) {
 	    property = NULL;
 	    result = XGetWindowProperty(winPtr->dispPtr->display, w, propName,
 		    0, 100000, False, XA_STRING, &actualType, &actualFormat,
@@ -2002,30 +2025,31 @@ TkpTestsendCmd(
 	    if (property != NULL) {
 		XFree(property);
 	    }
-	} else if (argv[4][0] == 0) {
+	} else if (Tcl_GetString(objv[4])[0] == 0) {
+            handler = Tk_CreateErrorHandler(winPtr->dispPtr->display,
+                    -1, -1, -1, NULL, NULL);
 	    XDeleteProperty(winPtr->dispPtr->display, w, propName);
+            Tk_DeleteErrorHandler(handler);
 	} else {
 	    Tcl_DString tmp;
 
 	    Tcl_DStringInit(&tmp);
-	    for (p = Tcl_DStringAppend(&tmp, argv[4],
-		    (int) strlen(argv[4])); *p != 0; p++) {
+	    for (p = Tcl_DStringAppend(&tmp, Tcl_GetString(objv[4]),
+		    (int) strlen(Tcl_GetString(objv[4]))); *p != 0; p++) {
 		if (*p == '\n') {
 		    *p = 0;
 		}
 	    }
-
+            handler = Tk_CreateErrorHandler(winPtr->dispPtr->display,
+                    -1, -1, -1, NULL, NULL);
 	    XChangeProperty(winPtr->dispPtr->display, w, propName, XA_STRING,
 		    8, PropModeReplace, (unsigned char*)Tcl_DStringValue(&tmp),
 		    p-Tcl_DStringValue(&tmp));
+            Tk_DeleteErrorHandler(handler);
 	    Tcl_DStringFree(&tmp);
 	}
-    } else if (strcmp(argv[1], "serial") == 0) {
+    } else if (index == TESTSEND_SERIAL) {
 	Tcl_SetObjResult(interp, Tcl_NewIntObj(localData.sendSerial+1));
-    } else {
-	Tcl_AppendResult(interp, "bad option \"", argv[1],
-		"\": must be bogus, prop, or serial", NULL);
-	return TCL_ERROR;
     }
     return TCL_OK;
 }
